/*
----------------------------------------------------------------------------

WOMBAT PI Metal detector
Arduino UNO R4 Minima Version

wombatpi.net

Last Modified:  21-Apr-2024


----------------------------------------------------------------------------
*/

#define VERSION "21APR2024"

#include "wombat.h"
#include "wombat_analog.h"
#include "target_sense.h"


#define BAUD_RATE (115200)




// raw analog samples to be averaged 
// Buffer SAMPLE_BUFFER_LENGTH long, contains a set of samples at different sample delays 'TIME_POINTS',
//
float samples[TIME_POINTS][SAMPLE_BUFFER_LENGTH];

// the last point / longest time on the discharge curve 
//
float sample_last [SAMPLE_BUFFER_LENGTH];

// For running averaging
//
double sums[TIME_POINTS] = {0.0};
double sum_last = 0.0;

double averages[TIME_POINTS] = {0.0};
double longAverages[TIME_POINTS] = {0.0};

double average_last = 0.0;
double mediumAverage_last = 0.0;
double longAverage_last = 0.0;

//double longAverageConduct[TIME_POINTS] = {0.0};
double averageConduct[TIME_POINTS] =  {0.0};

double normalised_average_curve[TIME_POINTS] = {0.0};
double normalised_curve[TIME_POINTS] = {0.0};

double signal_curve[TIME_POINTS] = {0.0};
double peak_signal_curve[TIME_POINTS] = {0.0};



double ratios[4]; 
int peakSampleCount = 0;  

#define MAX_TX_BUFFER (30)
char txBuffer[MAX_TX_BUFFER];


void setup() {  

  Serial.begin(BAUD_RATE);

  setupSample();
  setup_soundWave();    // Audio
  setupPulse();
   

  for(int i = 0 ; i < SAMPLE_BUFFER_LENGTH; i++)
  {
    // do nothing
  }

  

  // Not used
  //
  if(WIFI_SERIAL_ENABLED)
  {
    Serial1.begin(BAUD_RATE);
  }
}


#define ZERO (0)
#define ONE (1)
#define TWO (2)

// greater value means slower recovery, longer averaging
//
#define LONG_AVERAGE_FACTOR (4.0)


void loop() {  
  static int soundUpdateCount = 0;
  static int averageCount = 0;
  static double sum = 0.0;
  static int serialCheckCount = 0;   // check the serial port periodically
  static int printOutCount = 0;
  static int newAverageCounter;
  static double maxF;
  static int medianCounter = 0;
  static int loopCounter = 0;   

  int oldSample; 
  int index = 0;  
  double tempF ; 
  double signal;    
  
  if(sampleReady)
  {
    sampleReady = false;  
    
    // compensate_audio_for_pulse() ;  // compesnate for ~ 100Sec pulse
   
    // --running averages-- happens quckly and often (i.e at 500 Hz)
    // ! must be efficient !
    // replace the old value in the buffer with the new value
    // update the sum, re-calculate the average
    
    for(index = 0 ; index < TIME_POINTS; index++)
    {
      
      oldSample = samples[index][averageCount];           
      samples[index][averageCount] =( sampleArray[(*SAMPLE_INDEXES)[index]] );   // new raw sample from the set of samples, at the particular pulse width
      sums[index] -= oldSample;    // subtract old value from the sum
      sums[index] += samples[index][averageCount];  // add the new value to the sum

    }


    oldSample = sample_last[averageCount];
    tempF = 0;          
    sample_last[averageCount] = (sampleArray[SAMPLE_COUNT_MAX-1]);  
    sum_last -= oldSample;    // subtract old value from the sum
    sum_last += sample_last[averageCount];  // add the new value to the sum     

    // recalculate the running averages
    //    
    average_last = sum_last / SAMPLE_BUFFER_LENGTH;

    for(index = 0; index < TIME_POINTS; index++)
    {
      averages[index] = (sums[index] / SAMPLE_BUFFER_LENGTH);    
    }       

    printOutCount++;
    serialCheckCount++;
    soundUpdateCount++;
    
    averageCount++;   

    if(averageCount >= SAMPLE_BUFFER_LENGTH)
    {     
      // full set of samples complete
      //
      averageCount = 0;

      for(index = 0 ; index < TIME_POINTS; index++)
      {
        // re-calculate long averages
        //          
        tempF = (averages[index] -  longAverages[index]);
        tempF /= LONG_AVERAGE_FACTOR;          
        longAverages[index] += tempF;     

        // fast recovery, after the average has 'wound-up' due to being held on a target.
        // i.e The average follows the signal down quicker than it follows the target up.
        //
        if(tempF < 0)
        { 
           // add it again
           //
           longAverages[index] += tempF;    
        }
      }

      tempF = (average_last - longAverage_last);
      tempF /= LONG_AVERAGE_FACTOR;
      longAverage_last += tempF;     
    }
    else if (soundUpdateCount > SOUND_UPDATE_COUNT)
    {
      soundUpdateCount = 0;
      //do the local audio update 
      //        
      tempF = (averages[SOUND_INDEX] - longAverages[SOUND_INDEX]);  // strongest response for particular coil
      soundAlgorithm2(tempF);
      return;    
    }
    
    // Do our serial communication here if its time,   
    //          
    if (serialCheckCount > 100)    
    {
      serialCheckCount = 0;
      // check serial commands
      //   
      checkCommands();      
    }

    // Do our Discrimination and Target ID here if it's time
    //    
    // i.e If 600 Hz pulse rate, with 50-sample buffer, we will do this 600/50 = 12 times per second) 
    //
    else if (printOutCount  > SAMPLE_BUFFER_LENGTH)
    {      
      printOutCount = 0;
      // Time to send out serial port and do calculations
      //              
      if (mode == a) 
      { 
        for(int i = 0; i < TIME_POINTS; i++)
        {
          signal_curve[i] = averages[i]-longAverages[i];  
        }
        // signal ?
        //         
        tempF =  averages[0] - longAverages[0];   // Using sample at 18uSec
         
        if(tempF > THRESHOLD)                           // above our threshold ?
        {           
          TARGET_SENSE::targetID = NO_TARGET; // reset
          // while there is signal, store the ratios
          //                     
          if(tempF > maxF)
          {
            // going up (signal is increasing) , calculate ratios
            //
            maxF = tempF;
            ratios[R1] += signal_curve[1] / signal_curve[0];
            ratios[R2] += signal_curve[2] / signal_curve[0];
            ratios[R3] += signal_curve[4] / signal_curve[2];
            ratios[R4] += signal_curve[4] / signal_curve[0];
              
            peakSampleCount++;             
          }            
        }
        else
        {
          // if no signal and we have a new peak, 
          //
          if(maxF > 0.0)
          {                
            // average the ratios we were storing as the signal was increasing.
            //
            ratios[R1] /= peakSampleCount;
            ratios[R2] /= peakSampleCount;
            ratios[R3] /= peakSampleCount;
            ratios[R4] /= peakSampleCount;
            peakSampleCount = 0;
                         
            // Descrimination algorithm
            //
            DESCRIMINATION_ALGORITHM;
            
            //  Serial.println(TARGET_SENSE::targetID);

            for(int i = 0; i < 4; i++)
            {
              #ifdef USB_SERIAL_ENABLED
                Serial.print(ratios[i]);
                Serial.print(",");
              #endif

              ratios[i] = 0; // reset
            }  

            #ifdef USB_SERIAL_ENABLED
                Serial.print(averages[TIME_POINTS-1]);  // print the last raw reading for offset adjustment
                Serial.println();                
            #endif

            // We need a new peak before we come here again
            //
            maxF = 0.0;           
          }                      
        }                     
      }
    }     
  } 
}
